/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001, 2002 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 * "$Id:"
 */

#ifndef TVMET_UNARY_FUNCTIONAL_H
#define TVMET_UNARY_FUNCTIONAL_H

namespace tvmet {

/*
 *
 */
#define DECLARE_UNARY_FUNCTIONAL(OPNAME, OP)                   		\
template <class T>	                                 		\
struct OPNAME : public UnaryFunctional {                		\
  typedef bool						value_type;	\
  static inline value_type applyOn(T rhs) {		        	\
    return OP rhs;                                         		\
  }                                                         		\
  static void printOn(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #OPNAME << "<"	 			\
       << typeid(T).name() << ">"                         		\
       << std::endl;                          				\
  }     								\
};

DECLARE_UNARY_FUNCTIONAL(fcnl_LogicalNot, !)
#undef DECLARE_UNARY_FUNCTIONAL

/*
 *
 */
#define DECLARE_UNARY_FUNCTIONAL(OPNAME, OP)                   		\
template <class T>	                                 		\
struct OPNAME : public UnaryFunctional {                		\
  typedef T						value_type;	\
  static inline value_type applyOn(T rhs) {		        	\
    return OP rhs;                                         		\
  }                                                         		\
  static void printOn(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #OPNAME << "<"	 			\
       << typeid(T).name() << ">"                         		\
       << std::endl;                          				\
  }     								\
};

DECLARE_UNARY_FUNCTIONAL(fcnl_BitwiseNot, ~)
DECLARE_UNARY_FUNCTIONAL(fcnl_Negate, -)
#undef DECLARE_UNARY_FUNCTIONAL

/*
 *
 */
#define DECLARE_UNARY_FUNCTIONAL(FUNCNAME, MATHFUNC)             	\
template <class T>	                                 		\
struct FUNCNAME : public UnaryFunctional {                		\
  typedef T						value_type;	\
  static inline value_type applyOn(T rhs) {		        	\
    return MATHFUNC(rhs);               				\
  }                                                         		\
  static void printOn(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #FUNCNAME << "<"	 			\
       << typeid(T).name() << ">"                         		\
       << std::endl;                          				\
  }     								\
};

DECLARE_UNARY_FUNCTIONAL(fcnl_abs, std::abs)
DECLARE_UNARY_FUNCTIONAL(fcnl_cbrt, std::cbrt)
DECLARE_UNARY_FUNCTIONAL(fcnl_ceil, std::ceil)
DECLARE_UNARY_FUNCTIONAL(fcnl_floor, std::floor)
DECLARE_UNARY_FUNCTIONAL(fcnl_rint, std::rint)
DECLARE_UNARY_FUNCTIONAL(fcnl_sin, std::sin)
DECLARE_UNARY_FUNCTIONAL(fcnl_cos, std::cos)
DECLARE_UNARY_FUNCTIONAL(fcnl_tan, std::tan)
DECLARE_UNARY_FUNCTIONAL(fcnl_sinh, std::sinh)
DECLARE_UNARY_FUNCTIONAL(fcnl_cosh, std::cosh)
DECLARE_UNARY_FUNCTIONAL(fcnl_tanh, std::tanh)
DECLARE_UNARY_FUNCTIONAL(fcnl_asin, std::asin)
DECLARE_UNARY_FUNCTIONAL(fcnl_acos, std::acos)
DECLARE_UNARY_FUNCTIONAL(fcnl_atan, std::atan)
DECLARE_UNARY_FUNCTIONAL(fcnl_exp, std::exp)
DECLARE_UNARY_FUNCTIONAL(fcnl_log, std::log)
DECLARE_UNARY_FUNCTIONAL(fcnl_log10, std::log10)
DECLARE_UNARY_FUNCTIONAL(fcnl_sqrt, std::sqrt)

#ifdef HAVE_IEEE_MATH
DECLARE_UNARY_FUNCTIONAL(fcnl_asinh, ::asinh)
DECLARE_UNARY_FUNCTIONAL(fcnl_acosh, ::acosh)
DECLARE_UNARY_FUNCTIONAL(fcnl_atanh, ::atanh)
DECLARE_UNARY_FUNCTIONAL(fcnl_expm1, ::expm1)
DECLARE_UNARY_FUNCTIONAL(fcnl_log1p, ::log1p)
DECLARE_UNARY_FUNCTIONAL(fcnl_erf, ::erf)
DECLARE_UNARY_FUNCTIONAL(fcnl_erfc, ::erfc)
DECLARE_UNARY_FUNCTIONAL(fcnl_j0, ::j0)
DECLARE_UNARY_FUNCTIONAL(fcnl_j1, ::j1)
DECLARE_UNARY_FUNCTIONAL(fcnl_y0, ::y0)
DECLARE_UNARY_FUNCTIONAL(fcnl_y1, ::y1)
DECLARE_UNARY_FUNCTIONAL(fcnl_lgamma, ::lgamma)
#endif // HAVE_IEEE_MATH

#undef DECLARE_UNARY_FUNCTIONAL

/*
 *
 */
#define DECLARE_UNARY_TYPED_ARG_FUNCTIONAL(FUNCNAME, MATHFUNC, POD)    	\
template <>	                                 			\
struct FUNCNAME< POD > : public UnaryFunctional {           		\
  typedef POD						value_type;	\
  static inline value_type applyOn(POD rhs) {	        		\
    return MATHFUNC(rhs);               				\
  }                                                         		\
  static void printOn(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #FUNCNAME << "<"	 			\
       << typeid(POD).name() << ">"                         		\
       << std::endl;                          				\
  }     								\
};

DECLARE_UNARY_TYPED_ARG_FUNCTIONAL(fcnl_abs, std::labs, long int)
DECLARE_UNARY_TYPED_ARG_FUNCTIONAL(fcnl_abs, std::fabs, double)
#ifdef HAVE_LONG_DOUBLE
DECLARE_UNARY_TYPED_ARG_FUNCTIONAL(fcnl_abs, std::fabs, long double)
#endif // HAVE_LONG_DOUBLE

#undef DECLARE_UNARY_TYPED_ARG_FUNCTIONAL

/*
 *
 */
#define DECLARE_UNARY_TYPED_RET_FUNCTIONAL(FUNCNAME, MATHFUNC, POD)    	\
template <class T>                                 			\
struct FUNCNAME : public UnaryFunctional {           			\
  typedef T						value_type;	\
  static inline POD applyOn(T rhs) {		        		\
    return MATHFUNC(rhs);               				\
  }                                                         		\
  static void printOn(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #FUNCNAME << "<"	 			\
       << typeid(POD).name() << ">"                         		\
       << std::endl;                          				\
  }     								\
};

DECLARE_UNARY_TYPED_RET_FUNCTIONAL(fcnl_isnan, ::isnan, int)
DECLARE_UNARY_TYPED_RET_FUNCTIONAL(fcnl_isinf, ::isinf, int)
DECLARE_UNARY_TYPED_RET_FUNCTIONAL(fcnl_finite, ::finite, int)
#undef DECLARE_UNARY_TYPED_RET_FUNCTIONAL

#ifdef HAVE_COMPLEX_MATH
/*
 * complex<> types
 */
template <class T>                                 			\
struct FUNCNAME : public UnaryFunctional {           			\
  typedef T						value_type;	\
  static inline std::complex<T> applyOn(std::complex<T>& rhs) {		\
    return MATHFUNC(rhs);               				\
  }                                                         		\
  static void printOn(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #FUNCNAME << "<"	 			\
       << typeid(std::complex<T>).name() << ">"                		\
       << std::endl;                          				\
  }     								\
};

DECLARE_UNARY_TYPED_ARG_FUNCTIONAL(fcnl_abs, std::abs)
#undef DECLARE_UNARY_TYPED_RET_FUNCTIONAL

#endif // HAVE_COMPLEX_MATH



} // namespace tvmet

#endif // TVMET_UNARY_FUNCTIONAL_H

// Local Variables:
// mode:C++
// End:
